/*
CStringTree: Used to determine whether or not a unique string has been seen previously
Author: Jeremy Cattone, Adam King
Last Revised: 05/22/01
*/
#include "strtree.h"

#ifdef ENABLE_NODE_COUNTING
#pragma message ("### OPTION: Node Counting Enabled")
#else
#pragma message ("### OPTION: Node Counting Disabled")
#endif

#ifdef ENABLE_NODE_DATA
#pragma message ("### OPTION: Node Data Enabled")
#else
#pragma message ("### OPTION: Node Data Disabled")
#endif

// Ctor
CStringTree::CStringTree(bool bSort)
{
	// Prep the root node
	m_RootNode				= new SStringTreeNode;
	m_RootNode->NextNode	= NULL;
	m_RootNode->SubNode		= NULL;
	m_RootNode->StringData	= NULL;
	m_CachedNodes			= NULL;
	EnableSorting(bSort);

#ifdef ENABLE_NODE_COUNTING
	m_RootNode->iNodeCount	= 0;
#endif

#ifdef ENABLE_NODE_DATA
	m_RootNode->iNodeData	= 0;
#endif

#ifdef __WIN32__
	InitializeCriticalSection(&csCache);
	StrTreeCacheAllocations	= NULL;
	m_CachedNodes			= NULL;
	iAllocatedNodes			= 0;
	iCachedNodes			= 0;
	m_CachedNodes			= NULL;
	iDeletedNodes			= 0;
	iRecycledNodes			= 0;
#endif

}

// Dtor
CStringTree::~CStringTree()
{
	// Deallocate all memory
	Purge();

	_DeleteNode(m_RootNode);

#ifdef __WIN32__
	DeleteCriticalSection(&csCache);
#endif
}

// === Cached allocator support ===============================================
// The cached allocator will allocate a block of nodes when needed, and
// hand then out as needed.  This assists in preventing memory fragmentation
// and small-object losses when large numbers of nodes are being allocated
//
#ifdef __WIN32__
#ifdef USECACHEDALLOCATORS
#pragma message ("### OPTION: Cached Allocators Enabled")

// _AllocNode()
// A thread-safe node allocator that allocates new nodes in blocks of
// STRTREE_ALLOC_NODE_BLOCK_SIZE nodes at a time, and recycles deleted nodes
// when available.
//
SStringTreeNode* CStringTree::_AllocNode()
{
	// Thread-safe entry
	EnterCriticalSection(&csCache);

	// Check for deleted or cached nodes that may be recycled
	SStringTreeNode* NextAvailableNode = m_CachedNodes;
	if(NextAvailableNode)
	{
		--iCachedNodes;
		++iRecycledNodes;
		m_CachedNodes = m_CachedNodes->NextNode;
		LeaveCriticalSection(&csCache);
		return NextAvailableNode;
	}
	
	// else...
	// There were no cached nodes available, so let's allocate a new block
	NextAvailableNode = new SStringTreeNode[STRTREE_ALLOC_NODE_BLOCK_SIZE];
	iAllocatedNodes += STRTREE_ALLOC_NODE_BLOCK_SIZE - 1;

#ifdef _DEBUG
	// Allocation diagnostic information
	char szBuf[64];
	sprintf(szBuf, "StrTree: %d (%d total, %d deleted)\n", iAllocatedNodes, m_Nodes, iDeletedNodes);
	OutputDebugString(szBuf);
#endif

	// Record the block's primary pointer for deletion later
	STRTREEMEMBERALLOCATION *NextAvailableNodeAlloc = new STRTREEMEMBERALLOCATION;
	NextAvailableNodeAlloc->Allocated = NextAvailableNode;
	NextAvailableNodeAlloc->Next = StrTreeCacheAllocations;
	StrTreeCacheAllocations = NextAvailableNodeAlloc;
	iCachedNodes += STRTREE_ALLOC_NODE_BLOCK_SIZE - 1;
	
	// Initialize the new nodes.  Not strictly necessary, but assists in reducing errors
	for(int i = 0; i < STRTREE_ALLOC_NODE_BLOCK_SIZE; ++i)
	{
		NextAvailableNode[i].NextNode = m_CachedNodes;
		NextAvailableNode[i].StringData = NULL;
		NextAvailableNode[i].SubNode = NULL;
#ifdef ENABLE_NODE_COUNTING
		NextAvailableNode[i].iNodeCount = 0;
#endif
#ifdef ENABLE_NODE_DATA
		NextAvailableNode[i].iNodeData = 0;
#endif
		m_CachedNodes = &NextAvailableNode[i];
	}
	
	// Now that we know there are nodes available, grab a cached node and return
	NextAvailableNode = m_CachedNodes;
	m_CachedNodes = m_CachedNodes->NextNode;
	LeaveCriticalSection(&csCache);
	return NextAvailableNode;
}

// _FreeNode(...)
// Returns a node to the cache pool
//
void CStringTree::_FreeNode(SStringTreeNode* Node)
{
	EnterCriticalSection(&csCache);
	++iCachedNodes;
	++iDeletedNodes;
#ifdef ENABLE_NODE_COUNTING
	Node->iNodeCount = 0;
#endif
#ifdef ENABLE_NODE_DATA
	Node->iNodeData = 0;
#endif
	Node->NextNode = m_CachedNodes;
	m_CachedNodes = Node;
	LeaveCriticalSection(&csCache);
}

#else // USECACHEDALLOCATORS NOT defined
#pragma message ("### OPTION: Cached Allocators Disabled")

#define _AllocNode() new SStringTreeNode
#define _FreeNode(Node) if(Node) delete Node

#endif // USECACHEDALLOCATORS
#endif // __WIN32__
// === END Cached allocator support ===============================================

// EnableSorting(...)
// Dunno, can't figure it out ;-)
//
void CStringTree::EnableSorting(bool bSort)
{
	m_bEnableSorting = bSort;
}

// Purge: Removes all elements from the tree, and preps for new insertion 
// by creating a new (and empty) root node
//
void CStringTree::Purge()
{
	// Deallocate all memory, and reinitialize the root node
	SStringTreeNode *NextNode;
	while(m_RootNode)
	{
		NextNode = m_RootNode->NextNode;
		_DeleteNode(m_RootNode);
		m_RootNode = NextNode;
	}
	m_RootNode = _AllocNode();
	m_RootNode->NextNode = NULL;
	m_RootNode->SubNode = NULL;
	m_RootNode->StringData = NULL;
#ifdef ENABLE_NODE_COUNTING
	m_RootNode->iNodeCount = 0;
#endif
#ifdef ENABLE_NODE_DATA
	m_RootNode->iNodeData = 0;
#endif

	// Only initialize the cache if we're in a 32-bit environment
#ifdef __WIN32__
	STRTREEMEMBERALLOCATION* temp = StrTreeCacheAllocations;
	while(StrTreeCacheAllocations)
	{
		temp = StrTreeCacheAllocations->Next;
		delete StrTreeCacheAllocations;
		StrTreeCacheAllocations = temp;
	}
	StrTreeCacheAllocations = NULL;
	m_CachedNodes = NULL;
#endif
}

// AddString: Adds a string to the tree.  
// A return value of 'true' indicates that the string was not previously 
//   inserted into the tree, 
// A return value of 'false' indicates that the string already exists
//
bool CStringTree::AddString(const char* szString)
{
	SStringTreeNode *cursor = m_RootNode->SubNode;
	
	if(!szString)
		return false;
	
	// Handle the zero-length string
	if(!szString[0])
	{
		if(_Find(NULL, m_RootNode, true))
			return false;
		_AddSubNode(NULL, m_RootNode);
		return true;
	}

	// Attempt to add the string to each top level node
	while(cursor)
	{
		if(cursor && cursor->StringData && cursor->StringData[0] == szString[0])
		{
			// Yes, the node referenced by cursor has the same first character, so the link should be placed in that tree
			return _AddStringToNode(szString, cursor);
		}

		// Keep trying
		cursor = cursor->NextNode;
	}

	// If a suitable node was not found, then add a new node
	return _AddSubNode(szString, m_RootNode);//true;
}

// Display: Displays a rough graphical representation of the data organization to cout
//
void CStringTree::Display()
{
	DisplayToStream(cout);
}

// DisplayToStream: Displays a rough graphical representation of the data
// organization to the specified stream
//
void CStringTree::DisplayToStream(ostream &fout)
{
	_DisplayToStream(fout);
}

// Dump: Dumps all unique strings to cout (unsorted)
//
void CStringTree::Dump()
{
	DumpToFile("CON");
}

// DumpToFile: Dumps all unique strings to a file (unsorted)
//
void CStringTree::DumpToFile(char* szFileName)
{
	FILE *f = fopen(szFileName, "w");
	if(f)
	{
		_DumpToFile(f);
		fclose(f);
	}
}

// LoadFromFile: Load a list of strings from a file
//
void CStringTree::LoadFromFile(char* szFile)
{
	FILE *f = fopen(szFile, "r");
	if(f)
	{
		LoadFromFile(f);
		fclose(f);
	}
}

// LoadFromFile: Load a list of strings from a file
//
void CStringTree::LoadFromFile(FILE* fIn)
{
	char szBuffer[1024];
	while(!feof(fIn))
	{
		fgets(szBuffer, sizeof(szBuffer), fIn);
		AddString(szBuffer);
	}
}

// _AddSubNode: Adds szString string to Node as a subnode.  Note that the
// subnode is created unconditionally.  If you wish to perform a conditional
// or structure-preserving insertion, use _AddStringToNode
//
bool CStringTree::_AddSubNode(const char* szString, SStringTreeNode *Node)
{
	SStringTreeNode	*cursor = NULL;

	// Do not add a duplicate node
	cursor = _Find(szString, Node);
	if(cursor)
	{
#ifdef ENABLE_NODE_COUNTING
		++cursor->iNodeCount;
#endif
		return false;
	}

	// Create a new link in the current node to hold a pointer to the new node
	cursor = _AllocNode();
#ifdef ENABLE_NODE_COUNTING
	cursor->iNodeCount = 1;
#endif
	
	// Special case a NULL string, since NULL strings are valid data members
	if(szString && szString[0])
	{
		// The string is NOT null, allocate storage and copy the string data to the new node
		cursor->StringData = new char[strlen(szString) + 1];
		strcpy(cursor->StringData, szString);
	}
	else
	{
		// The string IS null, reflect this by nulling out the pointer to what would have been the string data
		cursor->StringData = NULL;
	}


	// Make sure the new node has no sub-nodes
	cursor->SubNode = NULL;

	if(m_bEnableSorting)
	{
		// Perform a linear probe on the existing subnodes to determine the proper insertion point
		SStringTreeNode	*Prev = Node;
		SStringTreeNode	*Next = Node->SubNode;
		
		while(	Next && 
				(((Next->StringData && cursor->StringData) && (Next->StringData[0] < cursor->StringData[0])) || (!Next->StringData))&& 
				cursor->StringData)
		{
			Prev = Next;
			Next = Next->NextNode;
		}

		if(Prev == Node)
			Prev->SubNode = cursor;
		else
			Prev->NextNode = cursor;

		cursor->NextNode = Next;
	}
	else
	{
		// Make the new link the head of the link list
		cursor->NextNode = Node->SubNode;

		// Attach the new node to the tree
		Node->SubNode = cursor;
	}

	// Done
	return true;
}

// _AddStringToNode: Here's the meat of the algorithm.  Searches for duplicates
// and manages the proper insertion of a new string into an existing node.
//
bool CStringTree::_AddStringToNode(const char* szString, SStringTreeNode *Node)
{
	SStringTreeNode	*link1 = NULL;

	// If szString is NULL, try to add a NULL subnode and return
	if(NULL == szString)		
		return _AddSubNode(NULL, Node);
	// else...

	// How many characters match, up to the length of the current string?
	for(unsigned int iCharactersMatched = 0; iCharactersMatched < strlen(Node->StringData); ++iCharactersMatched)
	{
		if(!Node->StringData[iCharactersMatched] || 
			szString[iCharactersMatched] != Node->StringData[iCharactersMatched])
		{
			break;
		}
	}
	
	// N == 0
	if(0 == iCharactersMatched)
	{
		// No characters match, we can't add the string to this node
		return false;
	}
	
	// N < length of this node's existing string - the string must be split
	if(iCharactersMatched < strlen(Node->StringData))
	{
		// Create temp Node 1: containing existing characters N+1 to end of current string
		link1 = _AllocNode();

		// Move all child nodes (except NULL) to Node 1
		link1->SubNode		= Node->SubNode;
		link1->NextNode		= NULL;

		// Copy the non-matching (trailing) characters to the new node
		link1->StringData	= new char[strlen(Node->StringData) - iCharactersMatched + 1];
#ifdef ENABLE_NODE_COUNTING
		link1->iNodeCount	= Node->iNodeCount;
#endif
#ifdef ENABLE_NODE_DATA
		link1->iNodeData	= Node->iNodeData;
#endif
		strcpy(link1->StringData, &Node->StringData[iCharactersMatched]);

		// Since we are migrating the node outward on the tree, migrate the node count as well
#ifdef ENABLE_NODE_COUNTING
		Node->iNodeCount	= 0;
#endif
#ifdef ENABLE_NODE_DATA
		Node->iNodeData	= 0;
#endif

		// Now attach the new node to the tree as a child of the current node
		Node->SubNode		= link1;

		// Truncate current node to matched characters only (and reallocate to save memory)
		char *temp = new char[iCharactersMatched + 1];
		strncpy(temp, Node->StringData, iCharactersMatched);
		temp[iCharactersMatched] = 0;
		char *temp2 = Node->StringData;
		Node->StringData = temp;

		// Deallocate the old string
		if(temp2)
			delete [] temp2;

		// Create temp Node 2
		// If the string is a substring of the existing node, then we simply add a NULL node,
		// since we've already split the node as needed
		// Otherwise (if the new string is longer than the original node's string), create a
		// new node that stores the remaining text, and make it a subnode of the current node.
		if(iCharactersMatched == strlen(szString))
		{
			_AddSubNode(NULL, Node);
		} else {
			_AddSubNode(szString + iCharactersMatched, Node);
		}

#ifdef ENABLE_NODE_COUNTING
		Node->iNodeCount = 1;
#endif
		return true;
	}

	// N == length of existing string
	// Are there any unmatched trailing characters left in the new string?
	if(iCharactersMatched < strlen(szString))
	{	// Yes
		// Are there existing subnodes
		if(Node->SubNode)
		{
			// Will any of the existing nodes accept the string?
			link1 = Node->SubNode;
			while(link1)
			{
				if(link1->StringData)
				{
					// A slight optimization when using a sorted list that should (on average) 
					// cut the comparisons in half, but will add a slight overhead in and of itself
					// NOTE: This causes the tree to report erroneous duplicates
					if(m_bEnableSorting && link1->StringData[0] > szString[iCharactersMatched])
						break;
					
					if(link1->StringData[0] == szString[iCharactersMatched])
					{
						// Yes, the node referenced by link1 has the same first character, so the link should be placed in that tree
						return _AddStringToNode(&szString[iCharactersMatched], link1);
					}
				}
				// Keep trying
				link1 = link1->NextNode;
			}

		}
		else
		{
			_AddSubNode(NULL, Node);
		}

		// No (no subnodes)
		return _AddSubNode(szString + iCharactersMatched, Node);
	}
	else
	{
		// If the node has sub-nodes, then make certain a NULL node is present, but if no sub-nodes exist, then the node
		// already represents the exact string being inserted
		if(Node->SubNode)
			return _AddSubNode(NULL, Node);
		else
		{
#ifdef ENABLE_NODE_COUNTING
			++(Node->iNodeCount);
#endif
		}
	}
	// No (there are no trailing characters left), so we have an exact match.  Duplicates are not allowed to be entered in the tree
	return false;
}

// _Find: Searches a node's immediate subnodes for existing data who's initial
// characters match those of the string in question
//
SStringTreeNode* CStringTree::_Find(const char* szString, SStringTreeNode *Node, bool bCaseSensitive)
{
	if(!Node)
		return NULL;
	SStringTreeNode* cursor = Node->SubNode;
	bool bMatch;
	while(cursor)
	{
		if(szString == NULL)
		{
			if(cursor->StringData == NULL)
				return cursor;
		}
		else
		{
			if(cursor->StringData)
			{
				if(bCaseSensitive)
					bMatch = (!cursor->StringData[0] && !szString[0]) || (cursor->StringData[0] != 0 && !strncmp(szString, cursor->StringData, strlen(cursor->StringData)));
				else
					bMatch = (!cursor->StringData[0] && !szString[0]) || (cursor->StringData[0] != 0 && !strnicmp(szString, cursor->StringData, strlen(cursor->StringData)));
				if(bMatch)
					return cursor;
			}

		}
		cursor = cursor->NextNode;
	}
	return NULL;
}

// _DeleteNode: Creates a new node (just kidding)
//
void CStringTree::_DeleteNode(SStringTreeNode *Node)
{
	if(!Node)
		return;

	// Find the first sub-node
	SStringTreeNode *SubNode = Node->SubNode;
	SStringTreeNode *temp;

	// Loop through and delete all sub-nodes
	while(SubNode)
	{
		temp = SubNode->NextNode;
		_DeleteNode(SubNode);
		SubNode = temp;
	}

	// Now delete the node itself
	if(Node->StringData)
		delete [] Node->StringData;
	_FreeNode(Node);
}

// SerializeToFile(...)
// Makes an attempt to serialize the binary data to a file, making no attempt to 
// maintain a human readable format
//
void CStringTree::SerializeToFile(char *szFileName)
{
	FILE *f = fopen(szFileName, "w+b");
	SerializeToFile(f);
	fclose(f);
}

// SerializeToFile(...)
// Makes an attempt to serialize the binary data to a file, making no attempt to 
// maintain a human readable format
//
void CStringTree::SerializeToFile(FILE *fOut)
{
	_SerializeToFile(fOut);
}

// _SerializeToFile(...)
// Recursive support routine for the public version of this function
//
void CStringTree::_SerializeToFile(FILE *fOut, SStringTreeNode* Node)
{
	SStringTreeNode* cursor;
	if(Node)
		cursor = Node;
	else
		cursor = m_RootNode;
	unsigned char szNodeEnd[1] = {255};
	unsigned char szLeafEnd[1] = {254};
	while(cursor)
	{
		if(Node && cursor->StringData)
			fwrite(cursor->StringData, 1, strlen(cursor->StringData), fOut);

		fwrite(szNodeEnd, 1, 1, fOut);

		if(cursor->SubNode)
			_SerializeToFile(fOut, cursor->SubNode);

		cursor = cursor->NextNode;
	}
	fwrite(szLeafEnd, 1, 1, fOut);
}


// _DumpToFile: Dumps the raw strings to the specified file
// Note that this function uses a recursive mode of operation
// to rebuild the strings, so an input set that varies widely
// in length, but little in character set may cause a stack overflow.
// This is, however, very unlikely unless the tree's depth is 
// exceptionally large
//
// NOTE: When in debug mode, the node count and data values are
// output along with the string data.  This will cause the output
// strings to differ from the actual data in the tree.  Take this 
// into account when debugging code that uses save / restore 
// operations of the tree
//
void CStringTree::_DumpToFile(FILE *fOut, SStringTreeNode* Node, char* prefix)
{
	SStringTreeNode* cursor;
	if(Node)
		cursor = Node;
	else
		cursor = m_RootNode;
	int temp;
	char* szBuildBuffer;
	while(cursor)
	{
		temp = 0;
		szBuildBuffer = NULL;
		if(prefix && prefix[0] != 0)
			temp = strlen(prefix);
		if(cursor->StringData && cursor->StringData[0] != 0)
			temp += strlen(cursor->StringData);
		 // +1 for terminating NULL, +1 for possible \n, and +1 in case the \n is converted to a CRLF for whatever reason
		szBuildBuffer = new char[temp + 3];
		szBuildBuffer[0] = 0;

		if(prefix && prefix[0])
			strcpy(szBuildBuffer, prefix);
		if(cursor->StringData && cursor->StringData[0])
			strcat(szBuildBuffer, cursor->StringData);
		if(Node && (!cursor->SubNode || !cursor->StringData))
		{
			fwrite(szBuildBuffer, 1, strlen(szBuildBuffer), fOut);
#ifdef _DEBUG
#pragma message ("### DEBUG mode is active, DumpToFile() and Dump() appending diagnostics information to each string.")
#ifdef ENABLE_NODE_COUNTING
			if(!cursor->SubNode)
			{
				char szBuildBuffer2[16];
				sprintf(szBuildBuffer2, " (%ld)", cursor->iNodeCount);
				fwrite(szBuildBuffer2, 1, strlen(szBuildBuffer2), fOut);
			}
#endif // ENABLE_NODE_COUNTING
#ifdef ENABLE_NODE_DATA
			if(!cursor->SubNode)
			{
				char szBuildBuffer2[50];
				sprintf(szBuildBuffer2, " (Data = %ld)", cursor->iNodeData);
				fwrite(szBuildBuffer2, 1, strlen(szBuildBuffer2), fOut);
			}
#endif // ENABLE_NODE_DATA
#endif // _DEBUG
			fwrite("\n", 1, 1, fOut);
		}

		if(cursor->SubNode)
			_DumpToFile(fOut, cursor->SubNode, szBuildBuffer);
		if(szBuildBuffer)
			delete [] szBuildBuffer;

		cursor = cursor->NextNode;
	}
}

// SaveToFile(...)
// Saves the contents of the tree to a file, dumping the raw input strings
// See the note preceeding the _DumpToFile() function
//
void CStringTree::SaveToFile(FILE* fhFile)
{
	_DumpToFile(fhFile);
}

// _DisplayToStream: Displays a rough graphical representation of the tree's
// data structure on the specified output stream
//
void CStringTree::_DisplayToStream(ostream &fout, SStringTreeNode* Node, int iLevel)
{
	SStringTreeNode* cursor;
	if(Node)
		cursor = Node;
	else
		cursor = m_RootNode;

	while(cursor)
	{
		for(int i = 0; i < iLevel; ++i)
			fout << "  ";
		fout << "+-";
		if(cursor->StringData)
			fout << cursor->StringData;
		else
			fout << "<NULL>";
		fout << endl;
		fout << flush;
		if(cursor->SubNode)
			_DisplayToStream(fout, cursor->SubNode, iLevel + 1);
		cursor = cursor->NextNode;
	}
}

// GetStatistics(...)
// Rudimentary stats gathering tool for diagnostics purposes
//
void CStringTree::GetStatistics(int *Nodes, int *Characters, int *NP, int *ST, int *FP)

{
	_GenerateStatistics();
	*Nodes = m_Nodes;
	*Characters = m_Characters;
	*NP = m_NullPointers;
	*ST = m_Strings;
	*FP = m_FullPointers;
}

// _GenerateStatistics(...)
// Helper for GetStatistics
//
void CStringTree::_GenerateStatistics(SStringTreeNode* Node)
{
	SStringTreeNode* cursor;
	if(Node)
		cursor = Node;
	else
	{
		m_Nodes = 0;
		m_Characters = 0;
		m_NullPointers = 0;
		m_Strings = 0;
		m_FullPointers = 0;
		cursor = m_RootNode;
	}
	m_Nodes++;

	while(cursor)
	{
		m_Nodes++;
		if(cursor->StringData)
		{
			m_Strings++;
			m_Characters += (strlen(cursor->StringData) + 1);
		}
		if(!cursor->SubNode)
			m_NullPointers++;
		else
			m_FullPointers++;
		
		if(!cursor->NextNode)
			m_NullPointers++;
		else
			m_FullPointers++;

		if(cursor->SubNode)
			_GenerateStatistics(cursor->SubNode);
		cursor = cursor->NextNode;
	}
}

// If MFC is in use, enable CStringArray I/O in the same
// manner as for files
#ifdef USE_MFC
bool CStringTree::AddCSArray(CStringArray *CSArray)
{
	bool bResult = true;
	for(int iIndex = 0; iIndex < CSArray->GetSize(); iIndex++)
		bResult = bResult & AddString((LPCTSTR) (CSArray->GetAt(iIndex)));
	return true;
}

void CStringTree::Dump(CStringArray *CSArray)
{
	_DumpToCSA(CSArray);
}

void CStringTree::_DumpToCSA(CStringArray *CSArray, SStringTreeNode* Node, char* prefix)
{
	SStringTreeNode* cursor;
	if(Node)
		cursor = Node;
	else
		cursor = m_RootNode;
	int temp;
	char* szBuildBuffer;
	while(cursor)
	{
		temp = 0;
		szBuildBuffer = NULL;
		if(prefix && prefix[0] != 0)
			temp = strlen(prefix);
		if(cursor->StringData && cursor->StringData[0] != 0)
			temp += strlen(cursor->StringData);
		 // +1 for terminating NULL, +1 for possible \n, and +1 in case the \n is converted to a CRLF for whatever reason
		szBuildBuffer = new char[temp + 3];
		szBuildBuffer[0] = 0;

		if(prefix && prefix[0])
			strcpy(szBuildBuffer, prefix);
		if(cursor->StringData && cursor->StringData[0])
			strcat(szBuildBuffer, cursor->StringData);
		if(Node && (!cursor->SubNode || !cursor->StringData))
		{
			CSArray->Add(szBuildBuffer);
		}

		if(cursor->SubNode)
			_DumpToCSA(CSArray, cursor->SubNode, szBuildBuffer);
		if(szBuildBuffer)
			delete [] szBuildBuffer;
			
		cursor = cursor->NextNode;
	}
}

#endif

// Find(...)
// Find the number of times a particular string has been added to the tree
//
unsigned long CStringTree::Find(const char* szString, bool bCaseSensitive)
{
	SStringTreeNode* Node = _SearchNode(szString, m_RootNode, bCaseSensitive);
	if(!Node)
		return NULL;
#ifdef ENABLE_NODE_COUNTING
	return Node->iNodeCount;
#else
	return Node != 0;
#endif
}

// FindNode(...)
// Finds a node by value
//
SStringTreeNode* CStringTree::FindNode(const char* szString, bool bCaseSensitive)
{
	return _SearchNode(szString, m_RootNode, bCaseSensitive);
}

// _SearchNode(...)
// Helper for FindNode - recursively searches a node for a match of the target string
//
SStringTreeNode* CStringTree::_SearchNode(const char* szString, SStringTreeNode* Node, bool bCaseSensitive)
{
	// If either of the parameters are NULL, fail immediately
	if(!Node || !szString)
		return NULL;

	// If the search string is a Null String, we match if the node has no string
	if(!szString[0] && (!Node->StringData && Node != m_RootNode))
		return Node;

	SStringTreeNode *Cursor = Node->SubNode;
	SStringTreeNode* matchNode;

	bool bMatched1 = false;
	bool bMatched2 = false;
	while(Cursor && !bMatched1 && !bMatched2)
	{
		// Match conditions:
		// 1) Zero-length string and Cursor has NULL StringData
		// 2) Non-zero length string and Cursor matches as many characters as Cursor->StringData, but less than length of szString
		// 3) Non-zero length string and Cursor matches as many characters as Cursor->StringData and szString
		bMatched1 =	/* [1] */	(!szString[0] && (!Cursor->StringData));
		bMatched2 = /* [2,3] */	(szString[0] && Cursor->StringData && !strnicmp(Cursor->StringData, szString, strlen(Cursor->StringData)));
		
		// Do nothing in the case where bMatched1 is true
		//if(bMatched1)
		//	;

		if(bMatched2)
		{
			// Check for the rest of the string
			if(Cursor->StringData && strlen(Cursor->StringData) < strlen(szString))
			{
				matchNode = _SearchNode(szString + strlen(Cursor->StringData), Cursor, bCaseSensitive);
				if(matchNode)
					return matchNode;
				bMatched2 = false;
			} else {
				matchNode = _SearchNode("", Cursor, bCaseSensitive);
				if(matchNode)
					return matchNode;
				bMatched2 = (!Cursor->SubNode);
			}
		}

		if(!bMatched1 && !bMatched2)
			Cursor = Cursor->NextNode;
	}
	if(Cursor)
		return Cursor;

	return NULL;
}

#ifdef ENABLE_EXTENDED_FUNCTIONS
#pragma message ("### OPTION: Extended Functions Enabled")
//------------------------------
#ifdef ENABLE_NODE_COUNTING
//------------------------------

	//------------------------------
	#ifdef ENABLE_NODE_DATA
	//------------------------------

	//===========================================================================
	// Counting = true, Data = true
	//===========================================================================

	//---------------------------------------------------------------------------
	bool CStringTree::AddStringEx(const char* szString, unsigned long count, NODE_DATA data)
	{
		SStringTreeNode *cursor = m_RootNode->SubNode;

		if(!szString)
			return false;

		// Handle the zero-length string
		if(!szString[0])
		{
			if(_Find(NULL, m_RootNode, true))
				return false;
			_AddSubNodeEx(NULL, m_RootNode, count, data);
			return true;
		}

		// Attempt to add the string to each top level node
		while(cursor)
		{
			if(cursor && cursor->StringData && cursor->StringData[0] == szString[0])
			{
				// Yes, the node referenced by cursor has the same first character, so the link should be placed in that tree
				return _AddStringToNodeEx(szString, cursor, count, data);
			}

			// Keep trying
			cursor = cursor->NextNode;
		}

		// If a suitable node was not found, then add a new node

		return _AddSubNodeEx(szString, m_RootNode, count, data);
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddSubNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count, NODE_DATA data)
	{
		SStringTreeNode	*cursor = NULL;

		// Do not add a duplicate node
		cursor = _Find(szString, Node);
		if(cursor)
		{
			cursor->iNodeCount += count;
			// Don't change data - assumes identical strings will reference identical
			// data values (also, if data holds a pointer to allocated memory, we'll
			// lose it otherwise)
			return false;
		}

		// Create a new link in the current node to hold a pointer to the new node
		cursor = _AllocNode();
		cursor->iNodeCount = count;
		cursor->iNodeData = data;

		// Special case a NULL string, since NULL strings are valid data members
		if(szString && szString[0])
		{
			// The string is NOT null, allocate storage and copy the string data to the new node
			cursor->StringData = new char[strlen(szString) + 1];
			strcpy(cursor->StringData, szString);
		}
		else
		{
			// The string IS null, reflect this by nulling out the pointer to what would have been the string data
			cursor->StringData = NULL;
		}


		// Make sure the new node has no sub-nodes
		cursor->SubNode = NULL;

		if(m_bEnableSorting)
		{
			SStringTreeNode	*Prev = Node;
			SStringTreeNode	*Next = Node->SubNode;

			while(	Next && 
					(((Next->StringData && cursor->StringData) && (Next->StringData[0] < cursor->StringData[0])) || (!Next->StringData))&& 
					cursor->StringData)
			{
				Prev = Next;
				Next = Next->NextNode;
			}

			if(Prev == Node)
				Prev->SubNode = cursor;
			else
				Prev->NextNode = cursor;

			cursor->NextNode = Next;
		}
		else
		{
			// Make the new link the head of the link list
			cursor->NextNode = Node->SubNode;

			// Attach the new node to the tree
			Node->SubNode = cursor;
		}

		// Done
		return true;
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddStringToNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count, NODE_DATA data)
	{
		SStringTreeNode	*link1 = NULL;

		// If szString is NULL, try to add a NULL subnode and return
		if(NULL == szString)
			return _AddSubNodeEx(NULL, Node, count, data);
		// else...

		// How many characters match, up to the length of the current string?
		for(unsigned int iCharactersMatched = 0; iCharactersMatched < strlen(Node->StringData); ++iCharactersMatched)
		{
			if(!Node->StringData[iCharactersMatched] ||
				szString[iCharactersMatched] != Node->StringData[iCharactersMatched])
			{
				break;
			}
		}

		// N == 0
		if(0 == iCharactersMatched)
		{
			// No characters match, we can't add the string to this node
			return false;
		}

		// N < length of this node's existing string - the string must be split
		if(iCharactersMatched < strlen(Node->StringData))
		{
			// Create temp Node 1: containing existing characters N+1 to end of current string
			link1 = _AllocNode();

			// Move all child nodes (except NULL) to Node 1
			link1->SubNode		= Node->SubNode;
			link1->NextNode	= NULL;

			// Copy the non-matching (trailing) characters to the new node
			link1->StringData	= new char[strlen(Node->StringData) - iCharactersMatched + 1];
			link1->iNodeCount	= Node->iNodeCount;
			link1->iNodeData	= Node->iNodeData;
			strcpy(link1->StringData, &Node->StringData[iCharactersMatched]);

			// Since we are migrating the node outward on the tree, migrate the node count as well
			Node->iNodeCount	= count;
			Node->iNodeData	= data;

			// Now attach the new node to the tree as a child of the current node
			Node->SubNode		= link1;

			// Truncate current node to matched characters only (and reallocate to save memory)
			char *temp = new char[iCharactersMatched + 1];
			strncpy(temp, Node->StringData, iCharactersMatched);
			temp[iCharactersMatched] = 0;
			char *temp2 = Node->StringData;
			Node->StringData = temp;

			// Deallocate the old string
			if(temp2)
				delete [] temp2;

			// Create temp Node 2
			// If the string is a substring of the existing node, then we simply add a NULL node,
			// since we've already split the node as needed
			// Otherwise (if the new string is longer than the original node's string), create a
			// new node that stores the remaining text, and make it a subnode of the current node.
			if(iCharactersMatched == strlen(szString))
			{
				_AddSubNodeEx(NULL, Node, count, data);
			} else {
				_AddSubNodeEx(szString + iCharactersMatched, Node, count, data);
			}

			return true;
		}

		// N == length of existing string
		// Are there any unmatched trailing characters left in the new string?
		if(iCharactersMatched < strlen(szString))
		{	// Yes
			// Are there existing subnodes
			if(Node->SubNode)
			{
				// Will any of the existing nodes accept the string?
				link1 = Node->SubNode;
				while(link1)
				{
					if(link1->StringData)
					{
						// A slight optimization when using a sorted list that should (on average)
						// cut the comparisons in half, but will add a slight overhead in and of itself
						// NOTE: This causes the tree to report erroneous duplicates
						if(m_bEnableSorting && link1->StringData[0] > szString[iCharactersMatched])
							break;

						if(link1->StringData[0] == szString[iCharactersMatched])
						{
							// Yes, the node referenced by link1 has the same first character, so the link should be placed in that tree
							return _AddStringToNodeEx(&szString[iCharactersMatched], link1, count, data);
						}
					}
					// Keep trying
					link1 = link1->NextNode;
				}

			}
			else
			{
				_AddSubNodeEx(NULL, Node, Node->iNodeCount, Node->iNodeData);
			}

			// No (no subnodes)
			return _AddSubNodeEx(szString + iCharactersMatched, Node, count, data);
		}
		else
		{
			// If the node has sub-nodes, then make certain a NULL node is present, but if no sub-nodes exist, then the node
			// already represents the exact string being inserted
			if(Node->SubNode)
				return _AddSubNodeEx(NULL, Node, count, data);
			else
			{
				Node->iNodeCount += count;
			}
		}
		// No (there are no trailing characters left), so we have an exact match.  Duplicates are not allowed to be entered in the tree

		return false;
	}

	//------------------------------
	#else	//ENABLE_NODE_DATA
	//------------------------------

	//===========================================================================
	// Counting = true, Data = false
	//===========================================================================

	//---------------------------------------------------------------------------
	bool CStringTree::AddStringEx(const char* szString, unsigned long count)
	{
		SStringTreeNode *cursor = m_RootNode->SubNode;

		if(!szString)
			return false;

		// Handle the zero-length string
		if(!szString[0])
		{
			if(_Find(NULL, m_RootNode, true))
				return false;
			_AddSubNodeEx(NULL, m_RootNode, count);
			return true;
		}

		// Attempt to add the string to each top level node
		while(cursor)
		{
			if(cursor && cursor->StringData && cursor->StringData[0] == szString[0])
			{
				// Yes, the node referenced by cursor has the same first character, so the link should be placed in that tree
				return _AddStringToNodeEx(szString, cursor, count);
			}

			// Keep trying
			cursor = cursor->NextNode;
		}

		// If a suitable node was not found, then add a new node

		return _AddSubNodeEx(szString, m_RootNode, count);
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddSubNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count)
	{
		SStringTreeNode	*cursor = NULL;

		// Do not add a duplicate node
		cursor = _Find(szString, Node);
		if(cursor)
		{
			cursor->iNodeCount += count;
			return false;
		}

		// Create a new link in the current node to hold a pointer to the new node
		cursor = _AllocNode();
		cursor->iNodeCount = count;

		// Special case a NULL string, since NULL strings are valid data members
		if(szString && szString[0])
		{
			// The string is NOT null, allocate storage and copy the string data to the new node
			cursor->StringData = new char[strlen(szString) + 1];
			strcpy(cursor->StringData, szString);
		}
		else
		{
			// The string IS null, reflect this by nulling out the pointer to what would have been the string data
			cursor->StringData = NULL;
		}


		// Make sure the new node has no sub-nodes
		cursor->SubNode = NULL;

		if(m_bEnableSorting)
		{
			SStringTreeNode	*Prev = Node;
			SStringTreeNode	*Next = Node->SubNode;

			while(	Next &&
					(((Next->StringData && cursor->StringData) && (Next->StringData[0] < cursor->StringData[0])) || (!Next->StringData))&&
					cursor->StringData)
			{
				Prev = Next;
				Next = Next->NextNode;
			}

			if(Prev == Node)
				Prev->SubNode = cursor;
			else
				Prev->NextNode = cursor;

			cursor->NextNode = Next;
		}
		else
		{
			// Make the new link the head of the link list
			cursor->NextNode = Node->SubNode;

			// Attach the new node to the tree
			Node->SubNode = cursor;
		}

		// Done
		return true;
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddStringToNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count)
	{
		SStringTreeNode	*link1 = NULL;

		// If szString is NULL, try to add a NULL subnode and return
		if(NULL == szString)
			return _AddSubNodeEx(NULL, Node, count);
		// else...

		// How many characters match, up to the length of the current string?
		for(unsigned int iCharactersMatched = 0; iCharactersMatched < strlen(Node->StringData); ++iCharactersMatched)
		{
			if(!Node->StringData[iCharactersMatched] ||
				szString[iCharactersMatched] != Node->StringData[iCharactersMatched])
			{
				break;
			}
		}

		// N == 0
		if(0 == iCharactersMatched)
		{
			// No characters match, we can't add the string to this node
			return false;
		}

		// N < length of this node's existing string - the string must be split
		if(iCharactersMatched < strlen(Node->StringData))
		{
			// Create temp Node 1: containing existing characters N+1 to end of current string
			link1 = _AllocNode();

			// Move all child nodes (except NULL) to Node 1
			link1->SubNode		= Node->SubNode;
			link1->NextNode	= NULL;

			// Copy the non-matching (trailing) characters to the new node
			link1->StringData	= new char[strlen(Node->StringData) - iCharactersMatched + 1];
			link1->iNodeCount	= Node->iNodeCount;
			strcpy(link1->StringData, &Node->StringData[iCharactersMatched]);

			// Since we are migrating the node outward on the tree, migrate the node count as well
			Node->iNodeCount	= count;

			// Now attach the new node to the tree as a child of the current node
			Node->SubNode		= link1;

			// Truncate current node to matched characters only (and reallocate to save memory)
			char *temp = new char[iCharactersMatched + 1];
			strncpy(temp, Node->StringData, iCharactersMatched);
			temp[iCharactersMatched] = 0;
			char *temp2 = Node->StringData;
			Node->StringData = temp;

			// Deallocate the old string
			if(temp2)
				delete [] temp2;

			// Create temp Node 2
			// If the string is a substring of the existing node, then we simply add a NULL node,
			// since we've already split the node as needed
			// Otherwise (if the new string is longer than the original node's string), create a
			// new node that stores the remaining text, and make it a subnode of the current node.
			if(iCharactersMatched == strlen(szString))
			{
				_AddSubNodeEx(NULL, Node, count);
			} else {
				_AddSubNodeEx(szString + iCharactersMatched, Node, count);
			}

			return true;
		}

		// N == length of existing string
		// Are there any unmatched trailing characters left in the new string?
		if(iCharactersMatched < strlen(szString))
		{	// Yes
			// Are there existing subnodes
			if(Node->SubNode)
			{
				// Will any of the existing nodes accept the string?
				link1 = Node->SubNode;
				while(link1)
				{
					if(link1->StringData)
					{
						// A slight optimization when using a sorted list that should (on average)
						// cut the comparisons in half, but will add a slight overhead in and of itself
						// NOTE: This causes the tree to report erroneous duplicates
						if(m_bEnableSorting && link1->StringData[0] > szString[iCharactersMatched])
							break;

						if(link1->StringData[0] == szString[iCharactersMatched])
						{
							// Yes, the node referenced by link1 has the same first character, so the link should be placed in that tree
							return _AddStringToNodeEx(&szString[iCharactersMatched], link1, count);
						}
					}
					// Keep trying
					link1 = link1->NextNode;
				}

			}
			else
			{
				_AddSubNodeEx(NULL, Node, Node->iNodeCount);
			}

			// No (no subnodes)
			return _AddSubNodeEx(szString + iCharactersMatched, Node, count);
		}
		else
		{
			// If the node has sub-nodes, then make certain a NULL node is present, but if no sub-nodes exist, then the node
			// already represents the exact string being inserted
			if(Node->SubNode)
				return _AddSubNodeEx(NULL, Node, count);
			else
			{
				Node->iNodeCount += count;
			}
		}
		// No (there are no trailing characters left), so we have an exact match.  Duplicates are not allowed to be entered in the tree

		return false;
	}

	//------------------------------
	#endif // ENABLE_NODE_DATA
	//------------------------------

//------------------------------
#else
//------------------------------

	//------------------------------
	#ifdef ENABLE_NODE_DATA
	//------------------------------

	//==============================================================================
	// Counting = false, Data = true
	//==============================================================================

	//---------------------------------------------------------------------------
	bool CStringTree::AddStringEx(const char* szString, NODE_DATA data)
	{
		SStringTreeNode *cursor = m_RootNode->SubNode;

		if(!szString)
			return false;

		// Handle the zero-length string
		if(!szString[0])
		{
			if(_Find(NULL, m_RootNode, true))
				return false;
			_AddSubNodeEx(NULL, m_RootNode, data);
			return true;
		}

		// Attempt to add the string to each top level node
		while(cursor)
		{
			if(cursor && cursor->StringData && cursor->StringData[0] == szString[0])
			{
				// Yes, the node referenced by cursor has the same first character, so the link should be placed in that tree
				return _AddStringToNodeEx(szString, cursor, data);
			}

			// Keep trying
			cursor = cursor->NextNode;
		}

		// If a suitable node was not found, then add a new node

		return _AddSubNodeEx(szString, m_RootNode, data);
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddSubNodeEx(const char* szString, SStringTreeNode *Node, NODE_DATA data)
	{
		SStringTreeNode	*cursor = NULL;

		// Do not add a duplicate node
		cursor = _Find(szString, Node);
		if(cursor)
		{
			// Don't change data - assumes identical strings will reference identical
			// data values (also, if data holds a pointer to allocated memory, we'll
			// lose it otherwise)
			return false;
		}

		// Create a new link in the current node to hold a pointer to the new node
		cursor = _AllocNode();
		cursor->iNodeData = data;

		// Special case a NULL string, since NULL strings are valid data members
		if(szString && szString[0])
		{
			// The string is NOT null, allocate storage and copy the string data to the new node
			cursor->StringData = new char[strlen(szString) + 1];
			strcpy(cursor->StringData, szString);
		}
		else
		{
			// The string IS null, reflect this by nulling out the pointer to what would have been the string data
			cursor->StringData = NULL;
		}


		// Make sure the new node has no sub-nodes
		cursor->SubNode = NULL;

		if(m_bEnableSorting)
		{
			SStringTreeNode	*Prev = Node;
			SStringTreeNode	*Next = Node->SubNode;

			while(	Next &&
					(((Next->StringData && cursor->StringData) && (Next->StringData[0] < cursor->StringData[0])) || (!Next->StringData))&&
					cursor->StringData)
			{
				Prev = Next;
				Next = Next->NextNode;
			}

			if(Prev == Node)
				Prev->SubNode = cursor;
			else
				Prev->NextNode = cursor;

			cursor->NextNode = Next;
		}
		else
		{
			// Make the new link the head of the link list
			cursor->NextNode = Node->SubNode;

			// Attach the new node to the tree
			Node->SubNode = cursor;
		}

		// Done
		return true;
	}





	//---------------------------------------------------------------------------
	bool CStringTree::_AddStringToNodeEx(const char* szString, SStringTreeNode *Node, NODE_DATA data)
	{
		SStringTreeNode	*link1 = NULL;

		// If szString is NULL, try to add a NULL subnode and return
		if(NULL == szString)
			return _AddSubNodeEx(NULL, Node, data);
		// else...

		// How many characters match, up to the length of the current string?
		for(unsigned int iCharactersMatched = 0; iCharactersMatched < strlen(Node->StringData); ++iCharactersMatched)
		{
			if(!Node->StringData[iCharactersMatched] ||
				szString[iCharactersMatched] != Node->StringData[iCharactersMatched])
			{
				break;
			}
		}

		// N == 0
		if(0 == iCharactersMatched)
		{
			// No characters match, we can't add the string to this node
			return false;
		}

		// N < length of this node's existing string - the string must be split
		if(iCharactersMatched < strlen(Node->StringData))
		{
			// Create temp Node 1: containing existing characters N+1 to end of current string
			link1 = _AllocNode();

			// Move all child nodes (except NULL) to Node 1
			link1->SubNode		= Node->SubNode;
			link1->NextNode	= NULL;

			// Copy the non-matching (trailing) characters to the new node
			link1->StringData	= new char[strlen(Node->StringData) - iCharactersMatched + 1];
			link1->iNodeData	= Node->iNodeData;
			strcpy(link1->StringData, &Node->StringData[iCharactersMatched]);

			// Since we are migrating the node outward on the tree, migrate the node count as well
			Node->iNodeData	= data;

			// Now attach the new node to the tree as a child of the current node
			Node->SubNode		= link1;

			// Truncate current node to matched characters only (and reallocate to save memory)
			char *temp = new char[iCharactersMatched + 1];
			strncpy(temp, Node->StringData, iCharactersMatched);
			temp[iCharactersMatched] = 0;
			char *temp2 = Node->StringData;
			Node->StringData = temp;

			// Deallocate the old string
			if(temp2)
				delete [] temp2;

			// Create temp Node 2
			// If the string is a substring of the existing node, then we simply add a NULL node,
			// since we've already split the node as needed
			// Otherwise (if the new string is longer than the original node's string), create a
			// new node that stores the remaining text, and make it a subnode of the current node.
			if(iCharactersMatched == strlen(szString))
			{
				_AddSubNodeEx(NULL, Node, data);
			} else {
				_AddSubNodeEx(szString + iCharactersMatched, Node, data);
			}

			return true;
		}

		// N == length of existing string
		// Are there any unmatched trailing characters left in the new string?
		if(iCharactersMatched < strlen(szString))
		{	// Yes
			// Are there existing subnodes
			if(Node->SubNode)
			{
				// Will any of the existing nodes accept the string?
				link1 = Node->SubNode;
				while(link1)
				{
					if(link1->StringData)
					{
						// A slight optimization when using a sorted list that should (on average)
						// cut the comparisons in half, but will add a slight overhead in and of itself
						// NOTE: This causes the tree to report erroneous duplicates
						if(m_bEnableSorting && link1->StringData[0] > szString[iCharactersMatched])
							break;

						if(link1->StringData[0] == szString[iCharactersMatched])
						{
							// Yes, the node referenced by link1 has the same first character, so the link should be placed in that tree
							return _AddStringToNodeEx(&szString[iCharactersMatched], link1, data);
						}
					}
					// Keep trying
					link1 = link1->NextNode;
				}

			}
			else
			{
				_AddSubNodeEx(NULL, Node, Node->iNodeData);
			}

			// No (no subnodes)
			return _AddSubNodeEx(szString + iCharactersMatched, Node, data);
		}
		else
		{
			// If the node has sub-nodes, then make certain a NULL node is present, but if no sub-nodes exist, then the node
			// already represents the exact string being inserted
			if(Node->SubNode)
				return _AddSubNodeEx(NULL, Node, data);
		}
		// No (there are no trailing characters left), so we have an exact match.  Duplicates are not allowed to be entered in the tree

		return false;
	}

	//------------------------------
	#endif //ENABLE_NODE_DATA
	//------------------------------

//------------------------------
#endif // ENABLE_NODE_COUNTING
//------------------------------

#else
#pragma message ("### OPTION: Extended Functions Disabled")
#endif
