// A string suffix-tree variant useful in matching previously seen strings
//
// Authored by Jeremy Cattone (Solitude042@hotmail.com)
// With much thanks to Adam King (agkking@netscape.net) for a significant contributions and review throughout the code
//
// Last Modified: 05/22/2001

#ifndef STRTREE_H
#define STRTREE_H

#ifdef WIN32
#define __WIN32__
#endif

#include <stdio.h>
#include <iostream.h>
#include <string.h>
#ifndef USE_MFC
#include <windows.h>
#else
#include <afxtempl.h>
#endif

#ifdef __BORLANDC__
#include <classlib\defs.h>
#endif

#define ENABLE_NODE_COUNTING
#ifdef __WIN32__
	#define USECACHEDALLOCATORS
#endif

//==============================
// These funcions allow you to associate any 32 bit data value with each string in the
// tree, and provides the following extra functions:
//   AddStringEx
//   _AddSubNodeEx (private)
//   _AddStringToNodeEx (private)
#define ENABLE_EXTENDED_FUNCTIONS
//==============================

const int STRTREE_ALLOC_NODE_BLOCK_SIZE = 1000;

//==============================
#ifdef ENABLE_EXTENDED_FUNCTIONS
	// Allows additional data item to be storred with each string.  I needed this for
	// my code to work, but you may want to keep it - it allows any additional data to
	// be stored or referenced by each string (eg: ptr, etc.)
	#define ENABLE_NODE_DATA

	#ifdef ENABLE_NODE_DATA
		typedef unsigned long	NODE_DATA;
	#endif
#endif
//==============================

struct SStringTreeNode
{
	char			*StringData;
	SStringTreeNode *NextNode;
	SStringTreeNode *SubNode;
#ifdef ENABLE_NODE_COUNTING
	unsigned long	iNodeCount;
#endif
#ifdef ENABLE_NODE_DATA
	NODE_DATA		iNodeData;
#endif
};

struct STRTREEMEMBERALLOCATION
{
	SStringTreeNode* Allocated;
	STRTREEMEMBERALLOCATION* Next;
};

class CStringTree
{
public:
	CStringTree(bool bSort = false); // ctor
	~CStringTree(); // dtor
	
	// Cleanup
	void Purge(); // removes all elements, and preps for new insertion

	// Interface
	bool AddString(const char* szString); // Adds a string to the tree.  A 'true' return indicates that the string was not previously in the tree, whereas 'false' indicates that the string already exists

//=== Alternate AddString implementations ======
#ifdef ENABLE_EXTENDED_FUNCTIONS
	#ifdef ENABLE_NODE_COUNTING
		#ifdef ENABLE_NODE_DATA
			// Add a string, and associate a count and data value with it.  If the string exists
			// the counts are added together, but the existing data value is left unchanged
			bool AddStringEx(const char* szString, unsigned long count, NODE_DATA data);
		#else
			// Add a string, and associate a count with it.  If the string exists
			// the counts are added together
			bool AddStringEx(const char* szString, unsigned long count);
		#endif
	#else
		#ifdef ENABLE_NODE_DATA
			// Add a string, and associate a data value with it.  If the string exists
			// the existing data value is left unchanged
			bool AddStringEx(const char* szString, NODE_DATA data);
		#endif
	#endif
#endif
//==============================================

	// Tree search tools
	unsigned long Find(const char* szString, bool bCaseSensitive = true);
	SStringTreeNode* FindNode(const char* szString, bool bCaseSensitive = true);

#ifdef USE_MFC
	bool AddCSArray(CStringArray *CSArray);
	void Dump(CStringArray *CSArray);
#endif

	// State management
	void Dump(); // Dumps all unique strings to cout (unsorted)
	void DumpToFile(char* szFileName); // Dumps all unique strings to a file (unsorted)
	void SaveToFile(FILE* fhFile); // Dumps all unique strings to a file (unsorted)

	void Display(); // Displays a rough graphical representation of the data organization to cout
	void DisplayToStream(ostream &fout = cout); // Displays a rough graphical representation of the data organization to the specified stream

	void LoadFromFile(char* szFileName); // Load a list of strings from a file
	void LoadFromFile(FILE* fIn); // Load a list of strings from a file

	void EnableSorting(bool bSort); // Enables / disables sorting
	void GetStatistics(int *Nodes, int *Characters, int *m_NullPointers, int *m_Strings, int *m_FullPointers); // Reports stats on the tree
	void SerializeToFile(char *szFileName); // Dumps binary data to a file
	void SerializeToFile(FILE *fOut); // Dumps binary data to a file

private:
	SStringTreeNode* _AllocNode();
	void			_FreeNode(SStringTreeNode* Node);

	bool _AddSubNode(const char* szString, SStringTreeNode *Node);
	bool _AddStringToNode(const char* szString, SStringTreeNode *Node);

	SStringTreeNode* _Find(const char* szString, SStringTreeNode *Node, bool bCaseSensitive = true);
	SStringTreeNode* _SearchNode(const char* szString, SStringTreeNode* Node, bool bCaseSensitive = true);

	void _SerializeToFile(FILE *fOut, SStringTreeNode* Node = 0);
	void _DisplayToStream(ostream &fout = cout, SStringTreeNode* Node = 0, int iLevel = 0);
	void _DumpToFile(FILE *fout, SStringTreeNode* Node = 0, char* prefix = 0);

	void _GenerateStatistics(SStringTreeNode* Node = NULL);

//==============================
#ifdef ENABLE_EXTENDED_FUNCTIONS
	#ifdef ENABLE_NODE_COUNTING
		#ifdef ENABLE_NODE_DATA
			bool _AddSubNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count, NODE_DATA data);
			bool _AddStringToNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count, NODE_DATA data);
		#else
			bool _AddSubNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count);
			bool _AddStringToNodeEx(const char* szString, SStringTreeNode *Node, unsigned long count);
		#endif
	#else
		#ifdef ENABLE_NODE_DATA
			bool _AddSubNodeEx(const char* szString, SStringTreeNode *Node, NODE_DATA data);
			bool _AddStringToNodeEx(const char* szString, SStringTreeNode *Node, NODE_DATA data);
		#endif
	#endif
#endif
//==============================

	// Variables
	void _DeleteNode(SStringTreeNode *Node);

	bool m_bEnableSorting;
	SStringTreeNode *m_RootNode;
	SStringTreeNode *m_CachedNodes;
	int m_Nodes;
	int m_NullPointers;
	int m_Strings;
	int m_FullPointers;
	int m_Characters;

#ifdef __WIN32__
	int iAllocatedNodes;
	int iCachedNodes;
	int iDeletedNodes;
	int iRecycledNodes;
	STRTREEMEMBERALLOCATION *StrTreeCacheAllocations;

	CRITICAL_SECTION csCache;
#endif


#ifdef USE_MFC
	void _DumpToCSA(CStringArray *CSArray, SStringTreeNode* Node = 0, char* prefix = 0);
#endif
};


#endif

